#!/usr/bin/python
# To run need schem_canv.py, dots.py, HelpFile, Component.so and Centroid.so
#
#   Calibration program that can be used to calibrate a camera anywhere using a 
#   calibration cube as detailed in dots.py.
#
# Usage: ./Calibrate.py <imagefile> <outfile>
#   where - imagefile is a frame of video containing the calibration cube
#         - outfile is the destination for the data necessary to create a calibration matrix
#
#  Krista Spence 2002/06

from Tkinter import *
from Dialog import Dialog
from schem_canv import schem_canv
import Image
import ImageTk
import Centroid
import Component #Author: Maxwell Sayles
import sys
import random

###############################################################################################
# Class: cam_im - inherits from Canvas
#   This canvas will display the imagefile passed to its constructor.
#   Upon creation, the class finds the centroids of all 'light colored'
#   connected components in the image.  These values are stored in self.centroids.
# Constructor Args:
#   parent - parent Tkinter widget
#   imfile - the image filename
#   schem_v - a reference to a schem_canv object (for calibration purposes, not always necessary)
###############################################################################################
class cam_im(Canvas):
    """Canvas will display image file passed in on the top right hand corner of the canvas"""
    def __init__(self, parent = None, imfile=None, schem_v = None):
        Canvas.__init__(self, parent, width=330, height=225, bd=2, relief=GROOVE,
                        cursor = 'draft_small')
        self.parent = parent
        self.schem = schem_v #reference to schematic canvas
        self.img = Image.open(imfile)
        self.tkim = ImageTk.PhotoImage(self.img)
        self.create_image( 0, 0, image=self.tkim, anchor=NW)
        self.bind('<Button-1>', self.select) 
        self.grid(row=0, column = 10, rowspan = 4, columnspan = 10, sticky=NSEW)
        self.fill_centroid_list()
                       
    def select(self, event): #called when a point on the canvas is selected by mouse click 
        x = self.canvasx(event.x) 
        y = self.canvasy(event.y)
                
        #unselect the dot on the schematic drawing and fill in 2d values chosen
        for face in self.schem.faces:
            for dot in face.dot_list:
                if dot.state == 1: #a dot has been selected in the schem drawing
                    centroid = self.find_centroid(x,y)
                    if centroid == None:
                        Dialog(self, title='Click Again', justify = 'center',
                               text='\t No centroid found.\n Please click closer to target location.',
                               bitmap = 'error', default=0, strings=('OK',))
                    else:  #unselect the dot on the schematic drawing and fill in 2d values
                           #for nearest centroid to selection
                        dot.fill2d(centroid[0], centroid[1])
                        dot.unselect()

    def find_centroid(self, x, y):
        """Given a point x,y, locate the nearest centroid on the cube and return.
           If no centroid, or >1 centroid is found within 5 pixels of the mouse click,
           None is returned."""
        closest = []
        
        for point in self.centroids:
            if (x > (point[0]-5) and x < (point[0]+5) and y > (point[1]-5) and y < (point[1]+5)):
                closest.append(point)
                
        if closest == [] or len(closest)>1:
            return None
        else:
            return closest[0] #return (x,y)
        
                    
    #Function: fill_centroid_list:
    #Calls functions from Centroid.so and Component.so
    def fill_centroid_list(self):
        """ Function: fill_centroid_list:
               - isolates all light colors in binary image (looking for dots on calibration box)
               - labels the connected components and finds the centroids of all objects of a
                 specified size - these values are stored in self.centroids"""
        
        imagestr = self.img.tostring()
        bin_imstr = ''
        #create a binary image: all light pixels from self.img are 1 in binary image
        for j in range(0,len(imagestr),3):
            if ((ord(imagestr[j])>100) and (ord(imagestr[j+1])>100)
                and (ord(imagestr[j+2])>80)):
                bin_imstr = bin_imstr + chr(1)
            else:
                bin_imstr = bin_imstr + chr(0)
        #bwim = Image.fromstring("L", self.img.size, bin_imstr)
        #bwim.save("testout.pgm") 
        lim = '0'*self.img.size[0]*self.img.size[1] #create buffer for labelled image
        nobj = Component.label(self.img.size, bin_imstr, lim, 8)
        #invocation is: get_centroids(dimensions, nobj, minsize, maxsize, img)
        self.centroids = Centroid.get_centroids(self.img.size, nobj, 4, 40, lim)
                   
###################################################################################       
# Class: control_pan - inherits from frame
#   Contains buttons and methods to save data, undo last step, get help and quit.
# Constructor Args:
#   parent - parent Tkinter widget
#   schem_v - reference to a schem_canv object
#   outfile - output file for calibration data
####################################################################################
class control_pan(Frame):
    def __init__(self, parent = None, schem_v = None, outfile = None):
        self.parent = parent
        self.schem = schem_v
        self.outf= outfile
        Frame.__init__(self, parent, relief=GROOVE, bd=2)
        self.grid(row=9, column=0, rowspan = 1, columnspan = 10, sticky=NSEW)
        Button(self, text='Save', command=self.save).grid(row=9, column = 1, columnspan = 2)
                                                                    
        Button(self, text='Quit', command=self.quit).grid(row=9, column = 3, columnspan = 2)
                                                     
        Button(self, text='Help', command=self.help).grid(row=9, column = 5, columnspan = 2)
                                                        
        Button(self, text='Clear Line', command=self.undo).grid(row = 9, column=7,
                                                           rowspan = 1, columnspan = 2)

    def notdone(self):
        Dialog(self, title='not yet implemented', text='not yet available',
               bitmap = 'error', default=0, strings=('OK',))

    def save(self): #write table data to file
        ans = Dialog(self, title='Verify save',
                     text='Confirm Save to "'+ self.outf + '"', bitmap='question',
                     default=1,
                     strings=('Yes', 'No'))
        if ans.num == 0:
            ofile = open(self.outf, 'w')
            for face in self.schem.faces:
                for dot in face.dot_list:
                    if dot.set == 1:
                        dot.writedata(ofile)
            ofile.close()

    def undo(self): #undo last selection
        if self.schem.last_selected != None:
            self.schem.last_selected.labels[3].configure(text = '')
            self.schem.last_selected.labels[4].configure(text = '')
            self.schem.last_selected.x2d = -1
            self.schem.last_selected.y2d = -1
            self.schem.last_selected.set = 0

    def help(self):
        new = Toplevel()
        Hfile = open('HelpFile', 'r')
        if Hfile == None:
            print 'didnt work'
        mtext = Hfile.read()
        Hfile.close()
        Message(new, text = mtext).pack()
        
    
    def quit(self):
        ans = Dialog(self, title='Verify quit',
                     text='Do you really want to quit?', bitmap='question',
                     default=1,
                     strings=('Yes', 'No'))
        if ans.num == 0:
            Frame.quit(self)

##########################################################################################################
# Class: view_chooser - inherits from Canvas
#    Small schematic drawing of the treadmill and possible camera positions.  Buttons used to choose
#    the camera position being viewed in the top right hand corner of the screen (on the cam_im Canvas).
# Constructor Args:
#    parent - parent Tkinter widget
#    schem_v - reference to a schem_canv object to contain the drawing of the calibration cube
##########################################################################################################
class view_chooser(Canvas):
    """Small schematic drawing of the treadmill and possible camera positions.  Buttons used to choose
         the camera position being displayed in the top right hand corner of the screen (on the cam_im Canvas)."""
    
    view_dict = { 'view1': ("diag", "X-"), 'view2' : ("diag", "Z+"), 'view3': ("diag", "X+"),
                       'view4' : ("diag", "Z-"), 'view5': ("straight", "X-"), 'view6': ("straight", "Z-"),
                       'view7': ("straight", "X+"), 'view8': ("straight", "Z+")}
    
    def __init__(self, parent=None, schem_v=None):
        self.schem_canv=schem_v
        Canvas.__init__(self, parent,width=330, height=225, relief=GROOVE, bd=2)
        #draw rectangle (treadmill)
        self.create_rectangle(135,100, 200, 125, fill='black', tags= ('tread',))
        #draw direction lines
        self.create_line(290,20 , 320,20 , 320,60, arrow='both', tags=('line',))
        self.create_text(285 ,20 , text = "Z", tags= ('label',))
        self.create_text(320, 70, text = "X", tags = ('label',))
        #draw arrows from buttons to 'treadmill'
        self.create_line(100, 72, 131,100 , arrow='last', tags= ('line',)) #view 1
        self.create_line(100, 110, 131, 110, arrow='last', tags= ('line',)) #view8
        self.create_line(100, 148, 131, 122, arrow='last', tags = ('line',)) #view2
        self.create_line(168, 72, 168, 98, arrow='last', tags = ('line',)) #view 5
        self.create_line(168, 148, 168, 125, arrow='last', tags = ('line',)) #view 7
        self.create_line(234, 72, 204, 100, arrow='last', tags = ('line',)) #view 3
        self.create_line(234, 110, 204, 110, arrow='last', tags = ('line',)) #view 6
        self.create_line(234, 148, 204, 122, arrow='last', tags = ('line',)) #view 4
        self.grid(row=5, column = 0, rowspan = 4, columnspan = 10, sticky=NSEW)

        #create buttons
        self.view1 = Button(self, text= "View 1", relief = RAISED,
                       command = lambda v ='view1': self.display_view(v))
        self.view1.place(relx=0.1, rely=0.2)
        self.view2 = Button(self, text="View 2", relief = RAISED,
                            command = lambda v ='view2': self.display_view(v))
        self.view2.place(relx=0.1, rely=0.65)
        self.view3 = Button(self, text="View 3", relief = RAISED,
                           command = lambda v ='view3': self.display_view(v) )
        self.view3.place(relx=0.7, rely=0.65)
        self.view4 = Button(self, text="View 4", relief = RAISED,
                            command = lambda v ='view4': self.display_view(v))
        self.view4.place(relx=0.7, rely=0.2)
        self.view5 = Button(self, text="View 5", relief = RAISED,
                            command = lambda v ='view5': self.display_view(v))
        self.view5.place(relx=0.4, rely=0.2)
        self.view6 = Button(self, text="View 6", relief = RAISED,
                            command = lambda v ='view6': self.display_view(v))
        self.view6.place(relx=0.7, rely=0.43)
        self.view7 = Button(self, text="View 7", relief = RAISED,
                            command = lambda v ='view7': self.display_view(v))
        self.view7.place(relx=0.4, rely=0.65)
        self.view8 = Button(self, text="View 8", relief = RAISED,
                            command = lambda v ='view8': self.display_view(v))
        self.view8.place(relx=0.1, rely=0.43)
      
     
    def display_view(self, view):
        #print view in NW corner of GUI
        self.schem_canv.delete_cube()
        self.schem_canv.draw_view(view_chooser.view_dict[view][0], view_chooser.view_dict[view][1])
        

##################################################
# MAIN PROGRAM
# Usage: ./Calibrate.py <imagefile> <outfile>
##################################################
if __name__ == '__main__':
    root = Tk()
    if len(sys.argv) != 3:
        print " Usage: ./Calibrate.py <imagefile> <outfile>"
        sys.exit(0)
    imfile = sys.argv[1]
      
    #create canvas to contain the schematic drawing of the cube 
    d_canv = schem_canv(parent = root)
    #create canvas containing frame of video
    photo_canv = cam_im(root, imfile, d_canv)
    #create canvas that allows user to choose which point of view they want
    view_chooser(root, d_canv)
    #create control panel
    control_pan(root, schem_v = d_canv, outfile = sys.argv[2])
    root.title('Camera Calibration')
    root.mainloop()
