/**
 *Centroid.c
 * Krista Spence 2002/06
 * 
 *Contains a function to be used in Python for finding the centroids
 *of objects within a grey scale image with connected components 
 *labelled.
 *
 *Compile with:
 *   gcc -Ipythonpath -c Centroid.c
 *   gcc -shared -o Centroid.so Centroid.o
 *
 * where 'pythonpath' is the path of the python include files
 *
 **/

#include <stdio.h>
#include <Python.h>
#include <string.h>

extern "C" void initCentroid(void);

PyObject* ErrorObject;
#define onError(message) \
        { PyErr_SetString (ErrorObject, message); return NULL; }


/**
 * this routine collects the size of every connected component of an image and 
 * then finds the centroids of the components whose size is witin a specified range
 *
 * invocation is: get_centroids(dimensions, nobj, minsize, maxsize, img)
 * where:
 *   dimensions is a 2-tuple of integers describing the image width and height
 *   nobj is the number of connected components found in the image
 *   minsize and maxsize specifies what the minimum and maximum size of an object
 *      must be to collect its centroid
 *   and img is a string represented an image with connected components labelled
 *
 * function will return a Python List containing tuples - (centroidxval, centroid yval)
 *
 **/

struct connect_comp{
  int size;
  int maxx;
  int minx;
  int maxy;
  int miny;
  double cx; /*centroid x val*/
  double cy; /*centroid y val*/
};

PyObject* get_centroids(PyObject* self, PyObject* args)
{
        /* vars for cmd line args*/
        int width;
        int height;
	int nobj;
	int minsize;
	int maxsize;
	char* img;
	int imgsize;
	int tempx;
	int tempy;
	int i = 0; /*counter*/
	PyObject* cent_list = PyList_New(0);
	PyObject* temp_list = PyList_New(0);
	
        if (!PyArg_ParseTuple (args, "(ii)iiis#", &width, &height, &nobj, &minsize, 
			       &maxsize, &img, &imgsize)) return NULL;
	
	connect_comp objs[nobj+1]; /*array of structs*/
	/*printf ("width = %i, height = %i\n", width, height);*/
	/*init structures */
	for (i = 0; i <= nobj; i++)
	  {
	    objs[i].size = 0;
	    objs[i].maxx = 0;
	    objs[i].maxy = 0;
	    objs[i].minx = width;
	    objs[i].miny = height;
	    objs[i].cx = 0;
	    objs[i].cy = 0;
	  }

	/*set vals for each connect_comp struct*/
	for (i = 0; i < imgsize; i++)
	  {
	    /*set current x, y vals*/
	    tempx = i%width;
	    tempy = i/width;
	    /*set min-max vals*/
	    if (tempx > objs[*img].maxx)
	      objs[*img].maxx = tempx;
	    else if (tempx < objs[*img].minx)
	      objs[*img].minx = tempx;
	    if (tempy > objs[*img].maxy)
	      objs[*img].maxy = tempy;
	    else if (tempy < objs[*img].miny)
	      objs[*img].miny = tempy; 

	    objs[*img].size++;
	    img++;
	  }
	
	for (i = 1; i <=nobj; i++)
	  {
	    temp_list = PyList_New(0);
	    
	    if ((objs[i].size <= maxsize) && (objs[i].size >= minsize))
	      { 
		objs[i].cx = (double(objs[i].maxx) + double(objs[i].minx))/2;
		objs[i].cy = (double(objs[i].maxy) + double(objs[i].miny))/2;
		PyList_Append(temp_list, PyFloat_FromDouble(objs[i].cx));
		PyList_Append(temp_list, PyFloat_FromDouble(objs[i].cy));
		PyList_Append(cent_list, PyList_AsTuple(temp_list));
		/*printf("object %i: size=%i Centroid=(%i, %i)\n", i, objs[i].size, objs[i].cx, objs[i].cy);*/
	      }
	  }

	/*printf("Length of cent_list %i\n", PyList_Size(cent_list));*/
        return cent_list;
}



/**
 * this is a list of module functions
 * these are functions that can be called by importing the module
 **/
typedef PyObject* PyFunction (PyObject*, PyObject*);

static struct PyMethodDef module_methods[] =
{
        { "get_centroids", get_centroids, METH_VARARGS },
	{ NULL, NULL , 0, NULL }
};


/**
 * initialization routine- called from import Centroid 
 **/
void initCentroid(void)
{
        PyObject* m;
        PyObject* d;

        m = Py_InitModule ("Centroid", module_methods);

        d = PyModule_GetDict(m);
        ErrorObject = (PyObject*)Py_BuildValue ("s", "Centroid.error");
        PyDict_SetItemString (d, "error", ErrorObject);
        if (PyErr_Occurred())
        {
                Py_FatalError ("can't initialize module Centroid");
        }
}






