/**
 *
 * Python extension for component labelling of binary images.
 *
 * Based on the octave module by Jeffrey E. Boyd.
 * Written by Maxwell Sayles, 2002.
 *
 *   - uses 4, 6, or 8 connectedness
 *   - See BKP Horn, Robot Vision, MIT Press, 1986, p 66 - 71
 *
 *   labeling scheme
 *
 *       +-+-+-+
 *       |D|C|E|
 *       +-+-+-+
 *       |B|A| |
 *       +-+-+-+
 *       | | | |
 *       +-+-+-+
 *
 *   A is the center pixel of a neighborhood.  In the 3 versions of
 *   connectedness:
 *
 *   4:  A connects to B and C
 *   6:  A connects to B, C, and D
 *   8:  A connects to B, C, D, and E
 *
 **/


#include <Python.h>

static PyObject* ErrorObject;
#define onError(message) \
        { PyErr_SetString (ErrorObject, message); return NULL; }

typedef unsigned char uchar;

static int width;
static int height;
static uchar* srcData;
static uchar* dstData;
static int connectedness;
static unsigned char labels [256]; /* maps index value to correct label value:  labelIndex = labels[labelIndex]; */
static unsigned char optimizedLabels [256];
static int objectCount;
static int x;
static int y;
static uchar* srcPixel;
static uchar* dstPixel;
static int B;
static int C;
static int D;
static int E;
static int temp;


static void do4Connectedness (void)
{       // apply 4 connectedness
        int min;

        if (B && C)
        {       // B and C are labeled
                if (B == C)
                        *dstPixel = B;
                else
                {       // B and C have different label values, use the lower label value
                        min = (B < C) ? B : C;
                        *dstPixel = min;
                        labels[B] = min;
                        labels[C] = min;
                }
        }
        else if (B)
        {       // B is object but C is not
                *dstPixel = B;
        }
        else if (C)
        {       // C is object but B is not
                *dstPixel = C;
        }
        else
        {       // B and C are not labeled, new object
                ++ objectCount;
                if (objectCount < 256)
                {
                        *dstPixel = objectCount;
                        labels[objectCount] = objectCount;
                }
        }
}


static void do6Connectedness (void)
{       // apply 6 connected ness
        int temp = 256; // no label with a value this high

        if (B || C || D)
        {       // atleast one labeled edge

                // find lowest value of the labeled edges
                if (B) temp = B;
                if (C) temp = (C < temp) ? C : temp;
                if (D) temp = (D < temp) ? D : temp;

                if (temp == 256) printf ("ComponentLabelling.do8Connectedness: ERROR!!!!");
                *dstPixel = temp; // label this pixel

                if (B) labels[B] = temp; // write lowest label value to referenced labels
                if (C) labels[C] = temp;
                if (D) labels[D] = temp;
        }
        else
        {       // no labeled edges, new object
                ++ objectCount;
                if (objectCount < 256)
                {
                        *dstPixel = objectCount;
                        labels[objectCount] = objectCount;
                }
        }
}


static void do8Connectedness (void)
{       // apply 8 connectedness
        int temp = 256; // no label with a value this high

        if (B || C || D || E)
        {       // atleast one labeled edge

                // find lowest value of the labeled edges
                if (B) temp = B;
                if (C) temp = (C < temp) ? C : temp;
                if (D) temp = (D < temp) ? D : temp;
                if (E) temp = (E < temp) ? E : temp;

                if (temp == 256) printf ("ComponentLabelling.do8Connectedness: ERROR!!!!");
                *dstPixel = temp; // label this pixel

                if (B) labels[B] = temp; // write lowest label value to referenced labels
                if (C) labels[C] = temp;
                if (D) labels[D] = temp;
                if (E) labels[E] = temp;
        }
        else
        {       // no labeled edges, new object
                ++ objectCount;
                if (objectCount < 256)
                {
                        *dstPixel = objectCount;
                        labels[objectCount] = objectCount;
                }
        }
}


/**
 * this routine performs the labelling of the image
 * invocation is: label (dimensions, src, dst, connectedness=6)
 * where:
 *   dimensions is a 2-tuple of integers describing the image width and height
 *   src and dst are both strings
 *   and connectedness may be either 4, 6, or 8
 *
 * this will handle a maximum of 255 objects
 * this is because an entry in a string can only contain values 0-255 and 0 is for the background
 **/
PyObject* CL_label (PyObject* self, PyObject* args)
{
        char* src;
        char* dst;
        int srcLength;
        int dstLength;

        /* get arguments from python tuple */
        connectedness = 6;
        if (!PyArg_ParseTuple (args, "(ii)s#s#|i",
                                &width, &height,
                                &src, &srcLength,
                                &dst, &dstLength,
                                &connectedness))
                return NULL;
        if ((connectedness != 4) && (connectedness != 6) && (connectedness != 8)) onError ("label: connectedness must be either 4, 6, or 8");
        if ((srcLength != width*height) || (dstLength != width*height)) onError ("label: vector lengths do not match image length");

        /* initialize all variables */
        objectCount = 0;
        memset (labels, 0, sizeof (labels));
        x = 0;
        y = 0;
        srcData = src;
        dstData = dst;
        srcPixel = src;
        dstPixel = dst;
        B = 0;
        C = 0;
        D = 0;
        E = 0;

        /* zero the destination buffer */
        memset (dst, 0, width*height);

        /* perform initial pass of labelling algorithm */
        for (y = 0;  y < height;  ++ y)
        {
                for (x = 0;  x < width;  ++ x)
                {
                        if (*srcPixel)
                        {       // this is a foreground pixel

                                // get the neighboring pixels B, C, D, and E
                                if (x == 0) B = 0; else B = labels[dstPixel[-1]];
                                if (y == 0) C = 0; else C = labels[dstPixel[-width]];
                                if (y == 0 || x == 0 ) D = 0; else D = labels[dstPixel[-width-1]];
                                if (y == 0 || x == width-1 ) E = 0; else E = labels[dstPixel[-width+1]];

                                if (connectedness == 4) do4Connectedness();
                                else if (connectedness == 6) do6Connectedness();
                                else if (connectedness == 8) do8Connectedness();
                                if (objectCount > 255)  x = width, y = height;
                        }
                        ++ srcPixel;
                        ++ dstPixel;
                }
        }


        /* consolidate image labels */
        for (x = 0;  x <= objectCount;  ++ x)
        {
                temp = x;
                while (labels[temp] != temp)
                {
                        temp = labels[temp];
                }
                labels[x] = temp;
        }


        /* optimize label table */
        objectCount = 0;
        for (x = 0;  x < 256;  ++ x)
        {
                if (labels[x] == x)
                {
                        optimizedLabels[x] = objectCount;
                        ++ objectCount;
                }
                else
                {
                        optimizedLabels[x] = optimizedLabels[labels[x]];
                }
        }
        -- objectCount; /* subtract 1 because the background does not count as an image */


        /* relabel objects with optimized label list */
        dstPixel = dstData;
        for (y = 0;  y < height;  ++ y)
        {
                for (x = 0;  x < width;  ++ x)
                {
                        *dstPixel = optimizedLabels [*dstPixel];
                        ++ dstPixel;
                }
        }

        return PyInt_FromLong (objectCount);
}



/**
 * list of module definitions
 **/
static struct PyMethodDef module_methods[] =
{
        { "label", CL_label, METH_VARARGS },
        { NULL, NULL , 0, NULL }
};


/**
 * initialization routine (called from 'import ComponentLabelling')
 **/
void initComponent (void)
{
        PyObject* m;
        PyObject* d;

        m = Py_InitModule ("Component", module_methods);

        d = PyModule_GetDict(m);
        ErrorObject = (PyObject*)Py_BuildValue ("s", "Component.error");
        PyDict_SetItemString (d, "error", ErrorObject);
        if (PyErr_Occurred())
        {
                Py_FatalError ("can't initialize module Component");
        }
}


