function Q = calibrate( match )
%
%   Q = calibrate( match ) - compute calibration matrix for a camera
%
%   match   -   table of point matches
%                   columns 1 thru 3 are X,Y,Z coords in world
%                   columns 4 and 5 are u,v coords in image
%
%   Q       -   calibration matrix
%

    N = size(match, 1);
    
    % Normalize calibration points
    sigma = std( match );
    avg = mean( match );
    T =     [   1/sigma(4)  0           -avg(4)/sigma(4);
                0           1/sigma(5)  -avg(5)/sigma(5);
                0           0           1                   ];
    U =     [   1/sigma(1)  0           0           -avg(1)/sigma(1);
                0           1/sigma(2)  0           -avg(2)/sigma(2);
                0           0           1/sigma(3)  -avg(3)/sigma(3);
                0           0           0           1                   ];

    X = U * [ match(:,1:3)'; ones(1,N) ];
    match(:,1:3) = X(1:3,:)';
    x = T * [ match(:,4:5)'; ones(1,N) ];
    match(:,4:5) = x(1:2,:)';
    
    % build matrix of calibration equations
    A = zeros( 2*N, 12 );
    A(1:2:2*N,1:3) = match(:,1:3);
    A(1:2:2*N,4) = 1;
    A(1:2:2*N,9)  = match(:,1) .* -match(:,4);
    A(1:2:2*N,10) = match(:,2) .* -match(:,4);
    A(1:2:2*N,11) = match(:,3) .* -match(:,4);
    A(1:2:2*N,12) = -match(:,4);

    A(2:2:2*N,5:7) = match(:,1:3);
    A(2:2:2*N,8) = 1;
    A(2:2:2*N,9)  = match(:,1) .* -match(:,5);
    A(2:2:2*N,10) = match(:,2) .* -match(:,5);
    A(2:2:2*N,11) = match(:,3) .* -match(:,5);
    A(2:2:2*N,12) = -match(:,5);
    
    % find solution, null space of A, using svd
    [u, s, v] = svd( A );
    
    s = diag(s(1:12,:));
    indx = find( s == min(s) );
    q = v(:,indx);
    
    Q = reshape(q, 4, 3 );
    Q = inv(T) * Q' * U;
endfunction
