#File contains: schem_canv class; cube_face class; dot class; data_table class
#Grouped together because a schem_canv has cube_faces, dots and a data_table
#For use with Calibrate.py
#Author: Krista Spence
#02/05

from Tkinter import *

import dots #'dots' contains the 3D world locations for each dot on the calibration cube

###################################################################
#Class: schem_canv
#    Contains a schematic drawing of a camera calibration cube
###################################################################
class schem_canv(Canvas):
    """Canvas to hold schematic drawing of calibration cube"""
    ref_dict = {"X-" : "Z+", "X+" : "Z-", "Z-" : "X-", "Z+" : "X+"}
    #given a key= the name for the left face of the cube, value = the corresponding right face of the cube (for a corner view only)

    def __init__(self, parent=None):
        Canvas.__init__(self,parent, width=330, height=225, bd=2, relief=GROOVE) #top left corner of window
        self.data_table = data_table(parent) #has a table with dot locations (3D and 2D)
        self.parent = parent
        self.grid(row=0, column = 0, rowspan = 4, columnspan = 10, sticky=NSEW)
        self.faces= [] #cube faces
        self.last_selected = None #last dot to have been selected on the canvas

    
    #Function: draw_view
    #Args: vp = viewpoint ('diag' means from corner, 'straight' means the cube is being viewed straight on)
    #      refview = reference for the top face of the cube to know what the world coordinates of each
    #                dot should be.  If in 'diag' view: will be the left face, else: will be the lower face.
    #Called when a viewpoint has been chosen - from an instance of the 'view-chooser' class
    def draw_view(self, vp, refview):  #vp = viewpoint, refview=reference point for top face of cube
        self.vp = vp
        self.refview = refview
        self.create_cube()  

    #Function: create_cube
    #Creates the necessary faces of the cube to be drawn
    def create_cube(self):  
        if self.vp == "diag": #diagonal view
            c1 = cube_face(self, ("L", self.refview)) #pass viewpoint (left,right, etc), and the cube face name
            self.faces.append(c1)
            c2 = cube_face(self, ("R", schem_canv.ref_dict[self.refview]))
            self.faces.append(c2)
            c3 = cube_face(self, ("T", "Y+"), self.refview)
            self.faces.append(c3)
        elif self.vp=="straight": #draw straight on view
            c1 = cube_face(self, ("F", self.refview))
            self.faces.append(c1)
            c2 = cube_face(self, ("U", "Y+"), self.refview)
            self.faces.append(c2)

    def delete_cube(self):
        for f in self.faces:
            f.delete()
        self.data_table.delete('data')
        self.data_table.y = 30
    
############################################################################################################
#Class: cube_face
# Constructor Args: canv = canvas the face is to be drawn on
#                   which_face = a tuple (side, label) where
#                                side- is the physical side to be drawn (L=left,R=right,T=top - 'diag' view,
#                                                                       U=upper,F=front - 'straight' view)
#                                label- is the 3D-Label for the edge (X+,Z- etc)
#                   Yref = the left face ('diag' view) or the front face ('straight' view)
############################################################################################################
class cube_face:
    def __init__(self, canv, which_face, Yref=None): 
        self.dvalues=dots.dots(which_face[1], Yref) #gets xyz values for each dot, currently in inches - from dots.py (ORDER MATTERS)
        self.which_face = which_face
        self.dot_list = []
        self.canvas = canv
        #draw the face with dots
        if self.which_face[0] == "L":
            self.canvas.create_polygon(70,70, 70, 170, 150, 210, 150, 110, fill='black', outline='white', tags=('side',))
            dot_loc=[(85, 150,'A'), (101, 154,'B'), (117, 162, 'C'), (133, 170,'D'),  #will write bottom to top
                     (85, 134,'E'), (101, 138,'F'), (117, 146,'G'), (133, 154,'H'),  #left to right, 3rd val is tag
                     (85, 116, 'I'), (101, 122, 'J'), (117, 130, 'K'), (133, 138, 'L'),
                     (85, 98, 'M'), (101, 106, 'N'), (117, 114, 'O'), (133,122, 'P')]
        elif self.which_face[0] == "T":
            self.canvas.create_polygon(70,70,150,110, 230,70, 150, 30, fill='black', outline='white', tags=('side',))
            dot_loc=[(100, 70, 'Q'), (115, 78, 'R'), (130, 86, 'S'), (145, 94, 'T'),    #will write top to bottom along
                     (115, 62, 'U'), (130, 70, 'V'), (145, 78, 'W'), (160, 86, 'X'),    #in \\\\ pattern, 3rd val is tag
                     (130, 54, 'Y'), (145, 62, 'Z'), (160, 70, 'AA'), (175, 78, 'AB'),
                     (145, 46, 'AC'), (160, 54, 'AD'), (175, 62, 'AE'), (190,70, 'AF')]
        elif self.which_face[0] == "R":
            self.canvas.create_polygon(150, 110,150,210, 230, 170, 230, 70, fill='black', outline='white', tags=('side',))
            dot_loc=[(165, 172, 'AG'), (180, 164, 'AH'), (195, 155, 'AI'), (210, 147, 'AJ'),  #will write bottom to top,
                    (165, 154, 'AK'), (180, 146, 'AL'), (195, 137, 'AM'), (210, 129, 'AN'),  # left to right, 3rd val is tag
                     (165, 138, 'AO'), (180, 128, 'AP'), (195, 119, 'AQ'), (210, 111, 'AR'),
                     (165, 120, 'AS'), (180, 110, 'AT'), (195, 101, 'AU'), (210, 91, 'AV')]
                            
        elif self.which_face[0] == "U":
            self.canvas.create_polygon(100, 110, 210, 110, 200, 20, 110, 20, fill='black', outline='white', tags=('side',))
            dot_loc=[(117,90 , 'Q'), (139, 90, 'R'), (161, 90, 'S'), (183, 90, 'T'), #will write bottom to top
                     (119, 70, 'U'), (140, 70, 'V'), (160, 70, 'W'), (181, 70, 'X'), #left to right, 3rd val is tag 
                     (121, 50, 'Y'), (141, 50, 'Z'), (159, 50, 'AA'), (179, 50, 'AB'),
                     (123, 30, 'AC'), (142, 30, 'AD'), (158, 30, 'AE'), (177,30, 'AF')]
                      

        elif self.which_face[0] == "F":
            self.canvas.create_polygon(100, 110, 210, 110, 210, 220, 100, 220, fill='black', outline='white', tags=('side',))
            dot_loc=[(117, 198,'A'), (139, 198,'B'), (161, 198, 'C'), (183, 198,'D'),  #will write bottom to top
                     (117, 176,'E'), (139, 176,'F'), (161, 176,'G'), (183, 176,'H'),  #left to right, 3rd val is tag
                     (117, 154, 'I'), (139, 154, 'J'), (161, 154, 'K'), (183, 154, 'L'),
                     (117, 132, 'M'), (139, 132, 'N'), (161, 132, 'O'), (183,132, 'P')]
        counter = 0
        for coord in self.dvalues: #create dots and assign 3D coordinates
            temp = dot(self.canvas, coord, dot_loc[counter]) #dot loc is coordinate location on canvas to draw the dot
            self.dot_list.append(temp)
            counter = counter+1

    def delete(self):
        self.canvas.delete('side')
        for dot in self.dot_list:
            self.canvas.delete(dot.tag)

            
#################################################################################################       
#Class: dot
#    Every dot knows whether it has been selected on the screen, if it is being pointed to, and
#    its 3D and 2D coordinates
#Constructor Args:
#   canv = the canvas the dot is to be drawn on
#   coord3d = a tuple (x,y,z)
#   coord2d = a tuple (x,y) where x and y are the 2D canvas coordinates the dot is to be drawn at
#################################################################################################
class dot:    
    def __init__(self,canv, coord3d, coord2d):
        self.canvas = canv
        self.state = 0 #selected?
        self.set = 0 #given u, v values?
        self.tag = coord2d[2]
        #draw the dot
        canv.create_oval(coord2d[0], coord2d[1], coord2d[0]+8 , coord2d[1]+8, fill='white', outline='black', tags= (self.tag,))
        self.x = coord3d[0]
        self.y= coord3d[1]
        self.z= coord3d[2]
        self.x2d = -1 #x 2d coord for dot
        self.y2d = -1 #y 2d coord for dot
        self.labels = self.canvas.data_table.fill_line(coord3d, self.tag) #fill in this dot's data line in the data_table (SE corner)
        canv.tag_bind(self.tag, '<Button-1>', self.select) 
        canv.tag_bind( self.tag, '<Enter>', self.enter) #highlight circle when entered
        canv.tag_bind( self.tag, '<Leave>', self.leave) #return circle to white

    def select(self, event):
        #un-highlight whatever dot was selected (and whatever line is currently selected)
        for face in self.canvas.faces:
            for dot in face.dot_list:
                if dot.state == 1:
                    dot.unselect()
        #highlight this dot
        self.canvas.itemconfig( self.tag, fill = 'red')

        #highlight all labels that go with this dot
        for label in self.labels:
            label.configure(bg = 'white')

        #jump view of data table so that selected line is second from top
        offset = (self.canvas.data_table.coords(self.tag)[1] - 30)/1470
        self.canvas.data_table.yview_moveto(offset)
        
        self.canvas.last_selected = self   
        self.state = 1 #selected
       

    def enter(self, event): 
        if self.state == 0:
            self.canvas.itemconfig( self.tag, fill='yellow' )

    def leave(self, event):
        if self.state == 0:
            self.canvas.itemconfig( self.tag, fill='white' )        

    def unselect(self):  #unselects the dot and the associated labels
        self.canvas.itemconfig(self.tag, fill='white')
        for label in self.labels:
            label.configure(bg = '#d5d6d5')
        self.state = 0
  
    def fill2d(self, x, y):   #fill in location on 2d picture for the dot
        self.labels[3].configure(text = x)
        self.labels[4].configure(text = y)
        self.x2d = x
        self.y2d = y
        self.set = 1

    def writedata(self, outfile): #writes x y z u v to file
        outfile.write( `self.x`+ ' ' + `self.y` + ' ' + `self.z` + ' ' + `self.x2d` + ' ' + `self.y2d` + '\n' )

##############################################################################################################
#Class: data_table
#       Displays a table of values: x,y,z are the 3D coord in real space; u,v are the 2D coord in the picture.
#       Manipulation is done from the schematic canvas.
##############################################################################################################
class data_table(Canvas):
    def __init__(self, parent = None):
        Canvas.__init__(self, parent, width = 330, height = 225, relief=GROOVE, bd = 2,
                        scrollregion = (0, 0, 330, 1470))
        self.vscroll = Scrollbar(parent, command=self.yview)
        self.configure(yscrollcommand=self.vscroll.set)
        self.grid(row=5, column = 10, rowspan = 5, columnspan = 10, sticky = NSEW)
        self.vscroll.grid(row = 5, column = 20, rowspan = 5, sticky = NS)
        #label columns
        headers = ['x', 'y', 'z', 'u', 'v']
        x = 0
        for title in headers:
            lbl = Label(self, text = title, padx = 28, pady=5, anchor = NW, relief = SOLID)
            self.create_window(x, 0, window = lbl, anchor = NW)
            x = x + 66
       
        self.y = 30 #used to place all additional labels on the canvas

      
    def fill_line(self, coord, tag): #tag will associate this row with a specific dot
        """Create a line in the data table holding xyz value information for one dot on the schematic drawing"""
        x = 0
        label_list = []
        win = None
        for val in coord: #fill in xyz vals for a dot
            temp = Label(self, text = '%2.4f'%val, padx = 5, pady = 10, anchor = NW, relief = SOLID, bd = 1, width = 66)
            self.create_window(x, self.y, window=temp, anchor = NW, tags = ('data', tag))
            label_list.append(temp)
            x = x + 66

        for n in range(2): #blank fields for camera x,y values
            temp = Label(self, text = '', padx = 15, pady = 10, anchor = NW, relief = SOLID, bd = 1, width = 66)
            win = self.create_window(x, self.y, window=temp, anchor = NW, tags = ('data', tag))
            label_list.append(temp)
            x = x + 66
        self.y = self.y + 30
        return label_list
